# Author: Jen Boulware, Yanlin Ye
# Common functions such as logging and temporary workspace for
# CSC's python geoprocessing/mapping scripts.
# Requirements: Spatial Analyst Extension

# Import arcpy, sys and traceback modules
import sys, traceback, os, errno, logging, logging.handlers, shutil
import ctypes.wintypes
import datetime, time
import arcpy
from arcpy import env
from arcpy.sa import *
from xml.dom.minidom import parse

# DEBUG mode: True for saving intermediate results for further investigation, False otherwise
# If the temporary workspace is initialized, all intermediate results will be saved there.
# Otherwise, the user's default file geodatabase will be used.
DEBUG = True

# Logger for logging
logger = logging.getLogger('csc.CscCommon')
__logHandlers = []

# Default document directory (for saving intermediate results, log file, etc)
# See getDocumentDir() and setDocumentDir() for details.
__documentDir = None

# Optional temporary geo-database for saving intermediate results
# See initTempWorkspace() and cleanTempWorkspace() for details.
__tempWorkspace = None
__prevWorkspace = None

# Count of temporary data items, used for constructing temporary data names
# See getTempName() for details.
__tmpCount = 0


class InputError(Exception):
    """
    Generic exception for user input errors. See error message for details.
    """
    pass

class NotTextNodeError:
    pass

def setDebugMode(debugMode=True):
    """
    Sets the debug mode, which controls the saving of intermediate results.
    See DEBUG for details.
    """
    global DEBUG
    DEBUG = debugMode

def initLogging(logfile, logger = None, level = logging.DEBUG):
    """
    Initializes logging using a simple log file. If the specified log file is relative,
    the current default document directory will be prepended.  A new log handler will be
    created and added to the specified logger, or the root logger if no logger is specified.

    The specified log file should be unregistered using the stopLogging(logger) method,
    which will unregister all previously registered log files for the same logger.
    """
    global __logHandlers

    # Rotating file handler
    file = logfile if os.path.isabs(logfile) else os.path.join(getDocumentDir(), logfile)
    handler = logging.handlers.RotatingFileHandler(file, 'a', 2*1024*1024, 1)
    format = '%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s'
    handler.setFormatter(logging.Formatter(format))
    if logger is None:
        logger = logging.getLogger()
    logger.addHandler(handler)
    logger.setLevel(level)
    __logHandlers.append((logger, handler))
    #logging.basicConfig(filename=file, \
    #                    level=logging.DEBUG, \
    #                    format='%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s')

    return

def stopLogging(logger = None):
    """
    Stops the logging by removing all log file registered by initLogging() for the same logger.
    If no logger is specified, the root logger will be used.
    """
    global __logHandlers

    if logger is None:
        logger = logging.getLogger()
    for i in range(len(__logHandlers)-1, -1, -1):
        (l, h) = __logHandlers[i]
        if l == logger:
            l.removeHandler(h)
            del __logHandlers[i]
    #del __logHandlers
    #__logHandlers = []

    return

def getDocumentDir():
    """
    Gets the default document directory under the user's "My Documents" folder for saving
    intermediate results, log files, etc. By default, the "NOAA Coastal Services Center"
    sub-directory under the current user's my documents folder will be returned.
    """
    global __documentDir
    if __documentDir is None:
        logger.debug("Default document directory not set.")
        setDocumentDir()
    return __documentDir

def setDocumentDir(path=None):
    """
    Sets the default document directory under the user's "My Documents" folder for saving
    intermediate results, log files, etc. By default, the "NOAA Coastal Services Center"
    sub-directory under the current user's my documents folder will be set as the document
    directory.
    If a relative path is specified, it will be concatenated to the end of the default
    directory. If an absolute path is specified, the absolute path will be set as the
    document directory.
    """

    global __documentDir

    logger.debug("Setting default document directory...")
    if (path is not None) and os.path.isabs(path):
        # User the specified absolute path
        logger.debug("Using the specified absolute path...")
        __documentDir = path
    else:
        # Gets the user's "My Documents" folder
        logger.debug("Detecting user's document folder...")
        buffer = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(0, 5, 0, 0, buffer)
        logger.debug("User's \"My Documents\": %s", buffer.value)

        # Constructs the default document directory
        __documentDir = os.path.join(buffer.value, 'NOAA Coastal Services Center')
        if path is not None:
            # Concatenates the specified relative path
            __documentDir = os.path.join(__documentDir, path)
            logger.debug("Specified relative path concatenated.")
        else:
            logger.debug("Path not specified.")
    logger.debug("Document directory: %s", __documentDir)

    # Tries to make the directory
    try:
        logger.debug("Trying to create it...")
        os.makedirs(__documentDir)
        logger.debug("Created successfully.")
    except OSError as e:
        if e.errno != errno.EEXIST:
            logger.exception("Failed to create directory: " + __documentDir)
            raise
        else:
            logger.debug("Directory existed.")
    return

def tryProcessing(function, *arguments, **options):
    """
    Tries to run the specified geoprocessing function with the supplied arguments inside
    a try-catch block for error logging and reporting through the addMessage(err,2) function.
    Logging and workspace will also be set up for the processing.

    This is a helper method designed for common tasks for running the script of a custom
    geoprocessing tool.

    Please note that the list of arguments should match those for the specified function
    if there are any.

    The method takes the following options:
        * name: Name of the specified geoprocessing functionality, used for logging and
          as default values for the following options. Default: "Analysis"
        * docdir: Subdirectory inside the "NOAA Coastal Services Center" folder in the user's
          "My Documents" folder for saving log files and the temporary workspace. Default: name
        * workspace: Temporary workspace for saving all intermediate results. Default: name + '.gdb'
        * logfile: Log file name. Default: name + '.log'

    If the function finishes successfully, its return value will be returned.  Otherwise,
    'None' will be returned instead.
    """
    # Encloses everything inside the following try-except block to capture, log, and report errors
    try:
        # Parses specified options
        name = options['name'] if ('name' in options) else 'Analysis'
        docdir = options['docdir'] if ('docdir' in options) else name
        workspace = options['workspace'] if ('workspace' in options) else name + '.gdb'
        logfile = options['logfile'] if ('logfile' in options) else name + '.log'

        # Sets up the default document directory
        setDocumentDir(docdir)

        # Sets up the logging
        initLogging(logfile)

        logger.debug("Name: %s, docdir: %s, workspace: %s, logfile: %s", name, docdir, workspace, logfile)
        logger.debug("Default document directory: %s", getDocumentDir())
        addMessage("Log file: " + os.path.join(getDocumentDir(), logfile))

        # Checks out the spatial analyst extension
        addMessage("Getting Spatial Analyst License...")
        if (arcpy.CheckExtension("Spatial") == "Available"):
            logger.debug("Spatial Analyst extension available. Checking out license...")
            arcpy.CheckOutExtension("Spatial")
        else:
            raise InputError("Spatial Analyst license is unavailable.")

        # Sets up the working environment, especially the temporary workspace (file geodatabase)
        addMessage("Setting up working environment...")
        initTempWorkspace(workspace)

        addMessage("Running %s..." % name)
        return function(*arguments)

    except InputError as e:
        logger.exception("An input validation error has occurred.")
        addMessage(str(e), 2)

    except Exception:
        #logger.exception("An error has occurred.")
        # Concatenates information together concerning the error into a message string
        #logger.debug("Preparing error details...")
        exc_type, exc_value, exc_traceback = sys.exc_info()
        msg = "An error has occurred.\n  " + "  ".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        arcpyMsg = arcpy.GetMessages(2)
        if arcpyMsg is not None and len(arcpyMsg) > 0:
            msg = msg + "ArcPy errors:\n" + arcpyMsg
        #logger.debug("Prepared error details: %s", msg)
        addMessage(msg, 2)

        # Sets the DEBUG mode so that the intermediate results won't be deleted
        setDebugMode(True)

    finally:
        # Checks in the Spatial Analyst extension
        addMessage("Checking in the Spatial Analyst extension...")
        arcpy.CheckInExtension("Spatial")

        # Cleans up the temporary workspace if necessary
        cleanTempWorkspace()

        # Done.
        logger.debug("Done.\n")

        # Stops logging to ensure that we remove the log handler previously registered by initLogging()
        stopLogging()
    return None

def addMessage(msg, severity=0):
    """
    Adds the specified message to geoprocessing messages as well as log it.
    """
    assert severity in (0, 1, 2), 'Invalid message severity (0 - message, 1 - warning, 2 - error): ' + str(severity)
    if severity == 0:
		logger.info(msg)
		arcpy.AddMessage(msg)
    elif severity == 1:
		logger.warning(msg)
		arcpy.AddWarning(msg)
    elif severity == 2:
		logger.error(msg)
		arcpy.AddError(msg)

def initTempWorkspace(workspaceName='TEMP.gdb'):
    """
    Initializes the temporary workspace by creating a file geodatabase within the default
    document directory and setting it as the ArcGIS' default workspace.  The current workspace
    will be saved and may be restored by calling method cleanTempWorkspace().
    """
    global __tempWorkspace, __prevWorkspace

    # Prepends the current time to the specified workspace name to make it unique
    workspaceName = time.strftime("%Y%m%dT%H%M%S") + '-' + workspaceName

    __tempWorkspace = os.path.join(getDocumentDir(), workspaceName)
    logger.debug("Temporary workspace: %s", __tempWorkspace)

    if os.path.exists(__tempWorkspace) and os.path.isdir(__tempWorkspace):
        logger.debug("Removing the existing file geodatabase...")
        shutil.rmtree(__tempWorkspace)

    logger.debug("Creating the temporary file geodatabase...")
    arcpy.CreateFileGDB_management(getDocumentDir(), workspaceName)

    __prevWorkspace = arcpy.env.workspace
    logger.debug("Saved previous workspace: %s", __prevWorkspace)

    arcpy.env.workspace = __tempWorkspace
    arcpy.env.overwriteOutput = True
    return

def cleanTempWorkspace():
    """
    Restores the previous workspace and cleans up the temporary workspace if necessary.
    This method will do nothing if no temporary workspace has been set up.  The temporary
    workspace, if set up, will not be cleaned and removed if it is in the debug mode.
    """
    global __tempWorkspace, __prevWorkspace
    # Cleans up the temporary workspace if necessary
    if __prevWorkspace is not None:
        logger.debug("Restoring workspace to: %s...", __prevWorkspace)
        arcpy.env.workspace = __prevWorkspace

        if DEBUG:
            addMessage("Intermediate results were saved in '%s'." % __tempWorkspace)
        else:
            addMessage("Cleaning up temporary workspace...")
            arcpy.Delete_management(__tempWorkspace)
    return

def getTempName(suffix, prefix = 'TMP'):
    """
    Gets the name of a temporary data item.
    """
    global __tmpCount
    __tmpCount += 1
    name = prefix + "_%02d__%s" % (__tmpCount, suffix)
    logger.debug("Temp name: %s", name)
    return name


def checkSpatialLic():
    # Check out the ArcGIS Spatial Analyst extension license
    if arcpy.CheckExtension("Spatial") == "Available":
        arcpy.CheckOutExtension("Spatial")
        return 0
    else:
        return 1


def checkPassedParamCount(cnt):
    # Since this is a common check, we pass the expected number
    #simple getArgumentCount tells us if they don't have enough
    #returns false if not, which raises the error in the main code
    if arcpy.GetArgumentCount() < cnt:
        return 1
    else:
        return 0


def getTextFromNode(node):
    """
    scans through all children of node and gathers the
    text. if node has non-text child-nodes, then
    NotTextNodeError is raised.
    """
    t = ""
    emptyNode = node.hasChildNodes()
    if emptyNode:
        for n in node.childNodes:
            if n.nodeType == n.TEXT_NODE:
                t += n.nodeValue
            else:
                raise NotTextNodeError
    else:
        t = "Null"
    return t

def nodeToDic(node):

    dic = {}
    multlist = {} # holds temporary lists where there are multiple children
    multiple = 0
    for n in node.childNodes:
        if n.nodeType != n.ELEMENT_NODE:
            continue

        # find out if there are multiple records
        if len(node.getElementsByTagName(n.nodeName)) > 1:
            multiple = 1
            # and set up the list to hold the values
            if not multlist.has_key(n.nodeName):
                multlist[n.nodeName] = []

        try:
            #text node
            text = getTextFromNode(n)
        except NotTextNodeError:
            if multiple:
                # append to our list
                multlist[n.nodeName].append(nodeToDic(n))
                dic.update({n.nodeName:multlist[n.nodeName]})
                continue
            else:
                # 'normal' node
                dic.update({n.nodeName:nodeToDic(n)})
                continue

        # text node
        if multiple:
            multlist[n.nodeName].append(text)
            dic.update({n.nodeName:multlist[n.nodeName]})
        else:
            dic.update({n.nodeName:text})
    return dic

def readXML(filename):
    dom = parse(filename)
    return nodeToDic(dom)

