# Author: Jen Boulware, Yanlin Ye
#
# Performs statistics regarding a reclassified raster using a zonal layer.
# The raster will be reclassified using a specified coefficient file and an
# optional population layer for determining the class of coefficients to use.
#
# Spatial Analyst Extension is required for running the script.

import logging, csv
import arcpy
from arcpy.sa import *
from CscCommon import *

# Indices of tool parameters, which should be consistent with the exact order
# of the parameters defined for the tool
INPUT_RASTER    = 0
ZONE_LAYER      = 1
ZONE_FIELD      = 2
COEF_FILE       = 3
COEF_CLASS      = 4
OUTPUT_LAYER    = 5
POP_LAYER       = 6
POP_FIELD       = 7
LOW_BREAKPOINT  = 8
HIGH_BREAKPOINT = 9

name = 'ISAT'
logger = logging.getLogger('csc.toolkit.ISAT')

# Gets the specified class of coefficients and load them from the coefficient file
def getRemapValue(coefficientFile, classToUse):
    """
    Gets the specified class of coefficients and load them from the coefficient file
    if they have not been loaded already.
    """

    # Opens the coefficient file
    logger.debug("Loading '%s' coefficients from file: %s...", classToUse, coefficientFile)
    f = open(coefficientFile, 'rt')
    coefficientData = csv.DictReader(f)

    # Checks the existence of required columns
    logger.debug("Checking required fields in the specified coefficient file...")
    for requiredField in ['Value', classToUse]:
        if not (requiredField in coefficientData.fieldnames):
            raise InputError("Missing required field '" + requiredField + "' in '" + coefficientFile + "'")

    # Builds the value list
    logger.debug("Building value list for reclassification...")
    valueList=[]
    for item in coefficientData:
        valueList.append([int(item["Value"]), int(float(item[classToUse]))])
    logger.debug("Value list: %s", str(valueList))

    # Saves the compiled value list
    rv = RemapValue(valueList)

    # Closes the coefficient file
    logger.debug("Closing the coefficient file...")
    f.close()

    # Returns the compiled value list
    return rv


# Reclassifies the cells within the specified group of population zones
def reclassifyCellsByGroup(inputRaster, groupClass, mask, coefficientFile, reclassOutput):
    """
    Reclassifies the cells within the specified group of population zones.
    """

    # First gets the polygon shapes and makes them a mask
    logger.debug("Applying mask for the population zones...")
    arcpy.env.mask = mask

    # Loads reclassification values from file
    logger.debug("Getting value list for reclassification...")
    remapValue = getRemapValue(coefficientFile, groupClass)

    # Reclassifies pixels in the population zone
    logger.debug("Reclassifying portion of the raster cells within the population zones...")
    groupReclassOutput = Reclassify(inputRaster, "Value", remapValue)
    if DEBUG:
        logger.debug("Saving intermediate results for reclassification of the population zones...")
        groupReclassOutput.save(getTempName("%s_RECLASS" % groupClass))

    # Removes the mask
    logger.debug("Removing mask for the population zones...")
    arcpy.env.mask = ""

    # Combines the pixels in the population zone with previously reclassified pixels
    logger.debug("Creating raster for detecting non-empty cells in the population zones...")
    isOutside = IsNull(groupReclassOutput)
    if DEBUG:
        logger.debug("Saving intermediate results for isNull raster...")
        isOutside.save(getTempName("%s_IS_OUTSIDE" % groupClass))
    logger.debug("Combining reclassified cells in the population zones with previously reclassified cells...")
    reclassOutput = Con(isOutside, reclassOutput, groupReclassOutput, "Value = 1")
    if DEBUG:
        logger.debug("Saving intermediate results for conditional merge raster...")
        reclassOutput.save(getTempName("%s_COMBINED" % groupClass))

    # Done.
    addMessage("Cells in the '%s' population zones reclassified." % groupClass)
    return reclassOutput

# Renames the specified fields to their new names, or removes fields with no specified new names
def renameFields(table, renames):
    """
    Renames the specified fields in a given table to their new names, or removes fields
    with no specified new names.  The fields to be renamed and their new names are
    specified as a dictionary from old names to new names.
    """

    # Makes a copy of the field mappings so that we can look up field names
    # in a case-insensitive way
    logger.debug("Translating field mappings to upper case names...")
    mappings = {}
    for name in renames.keys():
        mappings[name.upper()] = renames[name]

    # Maps the arcpy.Field types to geoprocessing types
    typeMappings = {"SmallInteger": "SHORT", "Integer": "LONG", "Single": "FLOAT",
                         "Double": "DOUBLE", "String": "TEXT", "Date": "DATE",
                         "OID": None, "Geometry": None, "Blob": "BLOB"}

    # Since there is no (easy) way of renaming fields using ArcPy, we need to
    # create a new field with the new name, copy values over, and delete the field
    # to be renamed.
    fields = arcpy.ListFields(table)
    for field in fields:
        name = field.name
        uname = name.upper()
        if mappings.has_key(uname) and mappings[uname] is not None:
            newName = mappings[uname]
            if not typeMappings.has_key(field.type) or typeMappings[field.type] is None:
                # Cloning a field is impossible for mismatched field types.
                raise Exception("No arcpy field type to geoprocessing field type mapping found " +
                                "for field '%s' of type '%s'." % (name, field.type))
            logger.debug("Copying field '%s' to '%s'...", name, newName)
            arcpy.AddField_management(table, newName, typeMappings[field.type], field.precision,
                                      field.scale, field.length, newName,
                                      field.isNullable, field.required, field.domain)
            arcpy.CalculateField_management(table, newName, "!" + name + "!", "PYTHON")

        if mappings.has_key(uname):
            logger.debug("Removing field '%s'...", name)
            arcpy.DeleteField_management(table, name)


# Saves statistical zones with valid statistics as the output data set
# and adds the statistical fields to the data set
def saveZonesWithStatistics(zoneLayer, zoneField, statTable, outputDataset):
    """
    Saves statistical zones with valid statistics as the output data set and adds the
    statistical fields to the data set.
    """

    # Makes a copy of the statistical zones with valid statistical results
    addMessage("Saving statistical zones with valid results as output feature class...")

    logger.debug("Making a temporary layer for the statistical zones...")
    tmpZonalLayer = getTempName("STAT_ZONES")
    arcpy.MakeFeatureLayer_management(zoneLayer, tmpZonalLayer)

    logger.debug("Joining the statistic results to the statistical zones...")
    arcpy.AddJoin_management(tmpZonalLayer, zoneField, statTable, zoneField.upper(), "KEEP_COMMON")

    logger.debug("Selecting all statistical zones with non-empty statistics...")
    arcpy.SelectLayerByAttribute_management(tmpZonalLayer, "NEW_SELECTION", "2 > 1")

    logger.debug("Removing the join of the statistic results...")
    arcpy.RemoveJoin_management(tmpZonalLayer, statTable)

    logger.debug("Saving selected statistical zones as output feature class...")
    tmpOutputDataset = getTempName("OUTPUT")
    arcpy.CopyFeatures_management(tmpZonalLayer, tmpOutputDataset)

    logger.debug("Deleting the temporary layer...")
    arcpy.Delete_management(tmpZonalLayer)

    # Adds wanted statistics to the output data set
    addMessage("Adding the statistical results to the output feature class...")

    logger.debug("Re-mapping the fields of the statistic table...")
    fields = {'AREA': 'AREA_IS', 'MEAN': 'MEAN_IS'}
    renameFields(statTable, fields)

    # Adds statistics fields to the zones
    logger.debug("Adding the statistical results to the statistical zones...")
    arcpy.JoinField_management(tmpOutputDataset, zoneField, statTable, zoneField.upper(),
        ['AREA_IS', 'MEAN_IS'])

    logger.debug("Copying the output feature class to specified location...")
    arcpy.CopyFeatures_management(tmpOutputDataset, outputDataset)

    if (not DEBUG):
        logger.debug("Removing temporary output data set...")
        arcpy.Delete_management(tmpOutputDataset)

# Performs statistics regarding a reclassified raster using a zonal layer.
def isat(inputRaster, zoneLayer, zoneField, coefficientFile, classToUse, outputDataset, \
            popLayer = None, popField = None, lowBreakpoint = None, highBreakpoint = None):
    """
    Performs statistics regarding a reclassified raster using a zonal layer and an optional
    population layer for determining the class of coefficient to use for each population zone.
    """

    logger.debug("Input raster: %s", inputRaster)
    logger.debug("Statistical zone layer: %s, zone field: %s", zoneLayer, zoneField)
    logger.debug("Coefficient file: %s, class to use: %s", coefficientFile, classToUse)
    logger.debug("Output data set: %s", outputDataset)

    # Performs the reclassification using the specified class of coefficients
    if (classToUse == "Calc"):
        lowBreakpoint = float(lowBreakpoint)
        highBreakpoint = float(highBreakpoint)
        logger.debug("Population layer: %s, population field: %s", popLayer, popField)
        logger.debug("Breakpoints: low: %f, high: %f", lowBreakpoint, highBreakpoint)
        if lowBreakpoint >= highBreakpoint:
            raise InputError("Invalid low/high break points!")

        # Groups the population zones based on the class of coefficients to use
        addMessage("Grouping population zones based on the breakpoints...")
        # Zone groups
        popZoneClasses = ["Low", "Med", "High"]
        logger.debug("Population zone classes: %s", str(popZoneClasses))

        # Counts
        popZoneCounts = [0, 0, 0]
        cursor = arcpy.SearchCursor(popLayer)
        for row in cursor:
            # Determines the class of coefficients to use for this population zone
            population = row.getValue(popField)
            if population < lowBreakpoint:
                # Low
                popZoneCounts[0] += 1
            elif population > highBreakpoint:
                # High
                popZoneCounts[2] += 1
            else:
                # Med
                popZoneCounts[1] += 1
        del row, cursor
        logger.debug("Population zone counts : %s", str(popZoneCounts))

        # Filters
        popField = arcpy.AddFieldDelimiters(popLayer, popField)
        logger.debug("Population field quoted: %s", popField)
        popZoneFilters = ("%(f)s < %(l)f;%(f)s >= %(l)f AND %(f)s <= %(h)f;%(f)s > %(h)f" % \
                          {"f": popField, "l": lowBreakpoint, "h": highBreakpoint}\
                          ).split(";")
        logger.debug("Population zone filters: %s", str(popZoneFilters))

        # Sets the output to the input raster with all cells set to NULL
        # just in case the population layer has a partial coverage over the input raster
        logger.debug("Creating the initial empty raster for combining reclassification results...")
        reclassOutput = SetNull(inputRaster, inputRaster, "Value >= 0 OR Value <= 0")
        if DEBUG:
            # Saves the initial null raster
            logger.debug("Saving initial empty raster for combining reclassification results...")
            reclassOutput.save(getTempName("RECLASS_Empty"))

        # Loops through the population zones by class of coefficients to use and processes
        # one set of population zones at a time
        for i in range(len(popZoneClasses)):
            addMessage("Population zone class: %s, feature count: %d. Reclassifying cells within zones..."\
                            % (popZoneClasses[i], popZoneCounts[i]), 0)
            if popZoneCounts[i] == 0:
                # Ignores class with no population zones
                addMessage("Ignored.")
            else:
                # Creates a temporary feature layer for this group of population zones for
                tmpPopLayer = getTempName("%s_POP_ZONES" % popZoneClasses[i])
                logger.debug("Creating temporary layer for this population zone group...")
                arcpy.MakeFeatureLayer_management(popLayer, tmpPopLayer, popZoneFilters[i])

                # Exports the population zones of this group as the mask data set
                mask = getTempName("%s_MASK" % popZoneClasses[i])
                logger.debug("Exporting the population zones as the mask data set...")
                arcpy.CopyFeatures_management(tmpPopLayer, mask)
                logger.debug("Deleting the temporary population layer...")
                arcpy.Delete_management(tmpPopLayer)

                # Performs reclassification for this zone group
                reclassOutput = reclassifyCellsByGroup(inputRaster, popZoneClasses[i], mask, coefficientFile, reclassOutput)

                # Removes the mask data set
                if not DEBUG:
                    logger.debug("Removing the mask layer...")
                    arcpy.Delete_management(mask)

    else:
        # OK so if they want to use a fixed class of coefficients (high,med,low),
        # we just build the remap values and do the reclassification
        logger.debug("Getting value list for reclassification...")
        remapValue = getRemapValue(coefficientFile, classToUse)

        # Reclassification
        addMessage("Reclassifying the input raster...")
        reclassOutput = Reclassify(inputRaster, "Value", remapValue)

    # In case we need to do attribute copying later
    #logger.debug("Building raster attribute table for the reclassified raster...")
    #arcpy.BuildRasterAttributeTable_management(reclassOutput, "NONE")
    if DEBUG:
        logger.debug("Saving the reclassified raster...")
        reclassOutput.save(getTempName("RECLASS_FINAL"))

    # Calculates statistics of the reclassified raster against the zone layer
    addMessage("Calculating statistics of the reclassified raster against the zone layer...")
    statTable = getTempName("STAT_TBL")
    ZonalStatisticsAsTable(zoneLayer, zoneField, reclassOutput, statTable, "DATA", "ALL")

    if (not DEBUG):
        # Deletes the reclassified raster
        logger.debug("Deleting the reclassified raster...")
        arcpy.Delete_management(reclassOutput)

    # Saves statistical zones with valid statistics as the output data set and
    # adds the statistical fields to the data set
    saveZonesWithStatistics(zoneLayer, zoneField, statTable, outputDataset)

    if (not DEBUG):
        # Deletes the table of statistical results
        logger.debug("Deleting the temporary table of statistical results...")
        arcpy.Delete_management(statTable)

    # Done.
    addMessage("%s finished successfully." % name)

if __name__ == "__main__":
    logger.debug("%s executed as a script.", name)
    # Runs analysis with all standard options for logging, temporary workspace, error handling, etc
    tryProcessing(isat, \
                  arcpy.GetParameterAsText(INPUT_RASTER), \
                  arcpy.GetParameterAsText(ZONE_LAYER), \
                  arcpy.GetParameterAsText(ZONE_FIELD), \
                  arcpy.GetParameterAsText(COEF_FILE), \
                  arcpy.GetParameterAsText(COEF_CLASS), \
                  arcpy.GetParameterAsText(OUTPUT_LAYER), \
                  arcpy.GetParameterAsText(POP_LAYER), \
                  arcpy.GetParameterAsText(POP_FIELD), \
                  arcpy.GetParameterAsText(LOW_BREAKPOINT), \
                  arcpy.GetParameterAsText(HIGH_BREAKPOINT), \
                  name=name, \
                  docdir='ISAT')

else:
    logger.debug("%s imported as a lib.", name)
