# Author: Yanlin Ye
#
# This is the validator class for the ISAT tool.
# The content of this file should be copied and pasted into the tool's
# validation script.
#  
# The file is saved here for easier editing in IDEs like Eclipse and 
# better source control.

import logging

# DEBUG mode. 
# Logging will be enabled if the DEBUG mode is True.
DEBUG = False

# Indices of tool parameters, which should be consistent with the exact order
# of the parameters defined for the tool
INPUT_RASTER    = 0
ZONE_LAYER      = 1
ZONE_FIELD      = 2
COEF_FILE       = 3
COEF_CLASS      = 4
OUTPUT_LAYER    = 5
POP_LAYER       = 6
POP_FIELD       = 7
LOW_BREAKPOINT  = 8
HIGH_BREAKPOINT = 9

logger = logging.getLogger('csc.toolkit.ISAT.Validator')

class ToolValidator:
    """Class for validating a tool's parameter values and controlling
    the behavior of the tool's dialog."""

    def __init__(self):
        """Setup arcpy and the list of tool parameters."""
        
        # Obtains parameter information
        import arcpy
        self.params = arcpy.GetParameterInfo()
        
        # Enables logging if in DEBUG mode
        if DEBUG:
            self.initLogging()
            
    def initLogging(self):
        """
        Initializes logging using a rotating file logger if it has not been initialized.
        """
        if len(logger.handlers) > 0:
            logger.debug("Handlers exist for logger: %d", len(logger.handlers))
            return
        
        import os, errno, ctypes.wintypes, logging.handlers
        
        # Gets the user's "My Documents" folder for saving log files
        buffer = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(0, 5, 0, 0, buffer)
        
        # Sets up the document directory
        toolkitDocDir = os.path.join(buffer.value, 'NOAA Coastal Services Center', 'ISAT')
        try:
            os.makedirs(toolkitDocDir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            
        logFile = os.path.join(toolkitDocDir, 'ISAT.Validator.log')
        
        # Simple file logger
        #logging.basicConfig(filename=logFile, \
        #                    level=logging.DEBUG, \
        #                    format='%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s')
        
        # Rotating file logger
        handler = logging.handlers.RotatingFileHandler(file, 'a', 2*1024*1024, 1)
        format = '%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s'
        handler.setFormatter(logging.Formatter(format))
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        logger.propagate = False
        
        return

    def initializeParameters(self):
        """Refine the properties of a tool's parameters.  This method is
        called when the tool is opened."""
        # Disables inputs that will depend on the input raster layer
        logger.debug("Initializing parameters...")
        self.params[POP_LAYER].enabled       = False
        self.params[POP_FIELD].enabled       = False
        self.params[LOW_BREAKPOINT].enabled  = False
        self.params[HIGH_BREAKPOINT].enabled = False
        return

    def updateMessages(self):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return
   
    def updateParameters(self):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""

        logger.debug("Updating parameters...")
        try:
            # Coefficient class to use
            logger.debug("Coefficient class to use: %s", self.params[COEF_CLASS].value)
            isCalc = (self.params[COEF_CLASS].value == 'Calc')
            logger.debug("Is 'Calc': %s", str(isCalc))
            
            self.params[POP_LAYER].enabled       = isCalc
            self.params[POP_FIELD].enabled       = isCalc
            self.params[LOW_BREAKPOINT].enabled  = isCalc
            self.params[HIGH_BREAKPOINT].enabled = isCalc

            logger.debug("Done updating parameters.")
            return
        except:
            logging.exception("Failed to update parameters.")
            raise
