# Author: Yanlin Ye
#
# Performs statistics regarding a reclassified raster using a zonal layer.
# The raster will be reclassified using a specified coefficient file and an
# optional population layer for determining the class of coefficients to use.
#
# The input raster can be optionally reclassified using a set of polygon layers.
# Cells within each polygon layer will be reclassified as the supplied new value
# classes.
#
# This is a convenient combination of the ISAT and Reclass by Polygons tools.
#
# Spatial Analyst Extension is required for running the script.

import logging, os, re
import arcpy
from arcpy.sa import *
from CscCommon import *
import ISAT, ReclassByPolygons

# Indices of tool parameters, which should be consistent with the exact order
# of the parameters defined for the tool
INPUT_RASTER   = 0
CLASS_NAME_FLD = 1
POLYGON_LAYER  = 2
RECLASS_AS     = 3
RECLASS_LIST   = 4
FILL_NODATA    = 5
#OUTPUT_RASTER  = 6

ZONE_LAYER      = 6
ZONE_FIELD      = 7
COEF_FILE       = 8
COEF_CLASS      = 9
OUTPUT_LAYER    = 10
POP_LAYER       = 11
POP_FIELD       = 12
LOW_BREAKPOINT  = 13
HIGH_BREAKPOINT = 14

name = 'ISAT with Reclass'
logger = logging.getLogger('csc.toolkit.ISATWithReclass')


def isatWithReclass(inputRaster, reclassifications, fillNodata, \
                    zoneLayer, zoneField, coefficientFile, classToUse, outputDataset, \
                    popLayer = None, popField = None, lowBreakpoint = None, highBreakpoint = None):
    """
    Performs statistics regarding a reclassified raster using a zonal layer and an optional
    population layer for determining the class of coefficient to use for each population zone.

    Optionally, reclassifies the specified input raster using the supplied polygon layers and their new
    value classes.  All meaningful fields will be copied over from the input raster layer
    to the output raster data set.  Symbology will be copied over as well.

    The reclassification rules should be in the format below.  The name of the new class is
    optional. The whole rule may be enclosed in a pair of single quotes.  Multiple rules
    should be separated by '; ' (a semicolon followed by a space).
        'Layer polygon-layer-name as new-value[: new-class]'
    """

    # Save the output
    addMessage("Reclassifying the input raster using the specified polygon layers...")
    outputRaster = None
    if reclassifications is not None and len(reclassifications) > 0:
        addMessage("Reclassification rules defined: %d" % (len(reclassifications)))
        outputRaster = getTempName("OUTPUT_RASTER")
        ReclassByPolygons.reclassByPolygons(inputRaster, reclassifications, fillNodata, outputRaster)
    else:
        addMessage("No reclassification rules are defined.")
        outputRaster = inputRaster

    addMessage("Running ISAT...")
    ISAT.isat(outputRaster, zoneLayer, zoneField, coefficientFile, classToUse, outputDataset, \
         popLayer, popField, lowBreakpoint, highBreakpoint)

    # Done.
    addMessage("%s finished successfully." % name)

if __name__ == "__main__":
    logger.debug("%s executed as a script.", name)
    # Runs analysis with all standard options for logging, temporary workspace, error handling, etc
    tryProcessing(isatWithReclass, \
                  arcpy.GetParameterAsText(INPUT_RASTER), \
                  arcpy.GetParameterAsText(RECLASS_LIST), \
                  ('true' == arcpy.GetParameterAsText(FILL_NODATA).lower()), \
                  arcpy.GetParameterAsText(ZONE_LAYER), \
                  arcpy.GetParameterAsText(ZONE_FIELD), \
                  arcpy.GetParameterAsText(COEF_FILE), \
                  arcpy.GetParameterAsText(COEF_CLASS), \
                  arcpy.GetParameterAsText(OUTPUT_LAYER), \
                  arcpy.GetParameterAsText(POP_LAYER), \
                  arcpy.GetParameterAsText(POP_FIELD), \
                  arcpy.GetParameterAsText(LOW_BREAKPOINT), \
                  arcpy.GetParameterAsText(HIGH_BREAKPOINT), \
                  name=name, \
                  docdir='ISAT')

else:
    logger.debug("%s imported as a lib.", name)
