# Author: Yanlin Ye
#
# This is the validator class for ISAT's "ISAT with Reclass" tool.
# The content of this file should be copied and pasted into the tool's
# validation script.
#  
# The file is saved here for easier editing in IDEs like Eclipse and 
# better source control.

import logging

# DEBUG mode. 
# Logging will be enabled if the DEBUG mode is True.
DEBUG = True

# Indices of tool parameters, which should be consistent with the exact order
# of the parameters defined for the tool
INPUT_RASTER   = 0
CLASS_NAME_FLD = 1
POLYGON_LAYER  = 2
RECLASS_AS     = 3
RECLASS_LIST   = 4
FILL_NODATA    = 5
#OUTPUT_RASTER  = 6

ZONE_LAYER      = 6
ZONE_FIELD      = 7
COEF_FILE       = 8
COEF_CLASS      = 9
OUTPUT_LAYER    = 10
POP_LAYER       = 11
POP_FIELD       = 12
LOW_BREAKPOINT  = 13
HIGH_BREAKPOINT = 14


logger = logging.getLogger('csc.toolkit.ISATWithReclass.Validator')

class ToolValidator:
    """Class for validating a tool's parameter values and controlling
    the behavior of the tool's dialog."""

    def __init__(self):
        """Setup arcpy and the list of tool parameters."""
        
        # Obtains parameter information
        import arcpy
        self.params = arcpy.GetParameterInfo()
        
        # Initializes cached raster data set name and class field name
        self.previousRasterLayer = None
        self.previousClassField  = None
        
        # Enables logging if in DEBUG mode
        if DEBUG:
            self.initLogging()
            
    def initLogging(self):
        """
        Initializes logging using a rotating file logger if it has not been initialized.
        """
        if len(logger.handlers) > 0:
            logger.debug("Handlers exist for logger: %d", len(logger.handlers))
            return
        
        import os, errno, ctypes.wintypes, logging.handlers
        
        # Gets the user's "My Documents" folder for saving log files
        buffer = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(0, 5, 0, 0, buffer)
        
        # Sets up the document directory
        toolkitDocDir = os.path.join(buffer.value, 'NOAA Coastal Services Center', 'ISAT')
        try:
            os.makedirs(toolkitDocDir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            
        logFile = os.path.join(toolkitDocDir, 'ReclassByPolygons.Validator.log')
        
        # Simple file logger
        #logging.basicConfig(filename=logFile, \
        #                    level=logging.DEBUG, \
        #                    format='%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s')
        
        # Rotating file logger
        handler = logging.handlers.RotatingFileHandler(logFile, 'a', 2*1024*1024, 1)
        fmt = '%(asctime)s [%(module)s %(lineno)3d] %(levelname)s: %(message)s'
        handler.setFormatter(logging.Formatter(fmt))
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        logger.propagate = False
        
        logger.debug("Logging configured. Log file: %s", logFile)
        return

    def initializeParameters(self):
        """Refine the properties of a tool's parameters.  This method is
        called when the tool is opened."""
        self.params[POP_LAYER].category       = "Population Layer and Breakpoints (Optional)"
        self.params[POP_FIELD].category       = "Population Layer and Breakpoints (Optional)"
        self.params[LOW_BREAKPOINT].category  = "Population Layer and Breakpoints (Optional)"
        self.params[HIGH_BREAKPOINT].category = "Population Layer and Breakpoints (Optional)"
        
        # Disables inputs that will depend on the input raster layer
        #logger.debug("Initializing parameters...")
        self.params[CLASS_NAME_FLD].enabled = False
        self.params[CLASS_NAME_FLD].value = None
        self.params[CLASS_NAME_FLD].filter.list = []
        self.params[POLYGON_LAYER].enabled = True
        self.params[RECLASS_AS].enabled = False
        self.params[RECLASS_AS].filter.list = []
        self.params[RECLASS_LIST].enabled = False
        self.params[RECLASS_LIST].value = None
        self.params[RECLASS_LIST].filter.list = []

        self.params[CLASS_NAME_FLD].category = 'Apply Change Scenarios (Optional)'
        self.params[POLYGON_LAYER].category  = 'Apply Change Scenarios (Optional)'
        self.params[RECLASS_AS].category     = 'Apply Change Scenarios (Optional)'
        self.params[RECLASS_LIST].category   = 'Apply Change Scenarios (Optional)'
        self.params[FILL_NODATA].category    = 'Apply Change Scenarios (Optional)'
        
        return
    
    def updatePopParamters(self):
        # Coefficient class to use
        logger.debug("Coefficient class to use: %s", self.params[COEF_CLASS].value)
        isCalc = (self.params[COEF_CLASS].value == 'Calc')
        logger.debug("Is 'Calc': %s", str(isCalc))
        
        self.params[POP_LAYER].enabled       = isCalc
        self.params[POP_FIELD].enabled       = isCalc
        self.params[LOW_BREAKPOINT].enabled  = isCalc
        self.params[HIGH_BREAKPOINT].enabled = isCalc
            
        return

    def updateMessages(self):
        """
        Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation.
        """

        if self.params[INPUT_RASTER].value is not None and len(self.params[RECLASS_AS].filter.list) == 0:        
            self.params[INPUT_RASTER].setErrorMessage('It seems that the specified input raster ' + \
                                                      'does not have an attribute table. Please ' + \
                                                      'create one for it and try again.')
        else:
            self.params[INPUT_RASTER].clearMessage()
        return
   
    def updateParameters(self):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""

        logger.debug("Updating parameters...")
        try:
            self.updatePopParamters()

            # Input raster layer
            inputRaster = self.params[INPUT_RASTER].value
            logger.debug("Raster layer: %s", str(inputRaster))
            
            # Updates the list of string fields for possible class names
            if (inputRaster is not None):
                if str(inputRaster) != self.previousRasterLayer:
                    logger.debug("New raster layer specified. Populating candidate class name fields...")
                    self.setFieldNames(self.params[INPUT_RASTER].value)
                    self.previousRasterLayer = str(inputRaster)
                else:
                    logger.debug("Raster layer not changed.")
            else:
                logger.debug("Raster layer not specified. Resetting parameters...")
                self.initializeParameters()

                logger.debug("Done updating parameters.")
                return

            # Class name field (optional)
            classField = self.params[CLASS_NAME_FLD].value
            logger.debug("Class name field: %s", str(classField))
            if classField != self.previousClassField:
                # Populates the list of value classes
                logger.debug("Class name field changed. Repopulating value classes...")
                self.setValueClasses(self.params[INPUT_RASTER].value, classField)
            elif (classField is None) and (len(self.params[RECLASS_AS].filter.list) == 0):
                # Populates the list of value classes even if the class name field is not specified
                logger.debug("Class name field not specified. Populating value classes for the first time...")
                self.setValueClasses(self.params[INPUT_RASTER].value, classField)
            else:
                logger.debug("Class name field not changed.")
            if len(self.params[RECLASS_AS].filter.list) == 0:
                logger.debug("Invalid input raster (likely missing attribute table).")
                logger.debug("Updating parameters stopped.")
                self.initializeParameters()
                return

            # Polygon layer
            polygonLayer = self.params[POLYGON_LAYER].value
            logger.debug("Polygon layer: %s", str(polygonLayer))
            
            # Enables/disables the options for reclassify the polygon layer
            if (polygonLayer is not None):
                logger.debug("Polygon layer selected. Enabling reclassification options...")
                self.params[RECLASS_AS].enabled = True

                # Reclassification options for the polygon layer
                reclassAs = self.params[RECLASS_AS].value
                logger.debug("Reclassify as: %s", str(reclassAs))
                if (reclassAs is not None) and (len(reclassAs) > 0):
                    logger.debug("Reclassification ready to be added.")
                    self.appendReclassification()
                else:
                    logger.debug("Reclassification option not selected.")
            else:
                logger.debug("Polygon layer not selected. Disabling reclassification options...")
                self.params[RECLASS_AS].enabled = False

            logger.debug("Done updating parameters.")
            return
        except:
            logging.exception("Failed to updating parameters.")
            raise

    def appendReclassification(self):
        """
        Appends the selected polygon layer and its reclassification option to the list.
        """
        
        # Composes the string representation of the reclassification
        layerName = str(self.params[POLYGON_LAYER].value)
        reclassKey = "Layer " + layerName + " as "
        reclassification = reclassKey + self.params[RECLASS_AS].value
        logger.debug("Reclassification: %s", reclassification)
        
        # Appends the reclassification to the filters
        reclasses = self.params[RECLASS_LIST].filter.list
        logger.debug("Adding reclassification to filters...")
        if reclasses is None:
            reclasses = []
        self.addItemToList(reclasses, reclassification, reclassKey)
        self.params[RECLASS_LIST].filter.list = reclasses
            
        logger.debug("New filters: %s", str(self.params[RECLASS_LIST].filter.list))
        
        # Appends it to the list of selected items, or replaces the existing row for 
        # the polygon layer if it has been previously added
        logger.debug("Adding reclassification to selected list...")
        reclasses = []
        if self.params[RECLASS_LIST].value is not None:
            reclasses = str(self.params[RECLASS_LIST].value).split(';')
        self.addItemToList(reclasses, "'" + reclassification + "'", reclassKey)
        self.params[RECLASS_LIST].value = '; '.join(reclasses)
        logger.debug("New selected list: %s", str(self.params[RECLASS_LIST].value))

        # Clears the polygon layer and its reclassification options
        logger.debug("Clearing the polygon layer's reclassification options...")
        self.params[RECLASS_AS].value = None
        #self.params[RECLASS_AS].enabled = False
        #self.params[POLYGON_LAYER].value = None

        # Enables the list of reclassifications
        self.params[RECLASS_LIST].enabled = True
        
        return
        
    def addItemToList(self, itemList, item, key):
        """
        Adds the specified item to the specified list.  If the item exists in the list,
        or more specifically, the item key (part of the item that can uniquely identify
        the item) can be found in any existing item, the first matching one will be 
        removed from the list.
        """
        
        logger.debug("Trying to add item to list...")
        logger.debug("Existing list: %s", str(itemList))
        logger.debug("Item: %s, key: %s", item, key)
        for i in itemList:
            if key in i:
                logger.debug("Found existing duplicate: %s", i)
                logger.debug("Removing it from list...")
                itemList.remove(i)
                break
        itemList.append(item)
        logger.debug("New list: %s", str(itemList))
        return itemList
        
    def setFieldNames(self, inputRaster):
        """
        Sets the list of string fields that may contain class names.
        """
        fields = []
        try:
            logger.debug("Getting fields for raster: %s", str(inputRaster))
            fields = [f.name for f in arcpy.ListFields(str(inputRaster), None, 'String')]
            logger.debug("Raster fields (strings only): %s", str(fields))
        except:
            logger.exception("Failed to set the list of candidate class name fields.")
            
        self.params[CLASS_NAME_FLD].filter.list = fields
        self.params[CLASS_NAME_FLD].enabled = (len(fields) > 0)
        self.previousClassField = None
        
        # Resets the list of value classes
        logger.debug("Resetting value classes...")
        self.params[RECLASS_AS].filter.list = []
        
    def setValueClasses(self, inputRaster, fldName):
        """
        Sets the list of all value classes for the specified raster layer and optional class field.
        If the class name field is specified, the classes will look like:
            '<class-value>: <class-name>'
        Otherwise, only the class value will be used.
        """
        vcs = []
        try:
            logger.debug("Getting value classes for raster: %s, field: %s", str(inputRaster), fldName)
            if (fldName is not None) and (len(fldName) > 0):
                cursor = arcpy.SearchCursor(inputRaster, "", "", "Value; " + fldName, "Value")
                vcs = [(str(r.getValue("Value")) + ": " + str(r.getValue(fldName))) for r in cursor]
            else:
                cursor = arcpy.SearchCursor(inputRaster, "", "", "Value", "Value")
                vcs = [str(r.getValue("Value")) for r in cursor]
            logger.debug("Classes: %s", str(vcs))
        except:
            logger.exception("Failed to query the list of value classes from the input raster.")
            self.params[RECLASS_AS].enabled = False
            
        self.params[RECLASS_AS].filter.list = vcs
