# Author: Jen Boulware, Yanlin Ye
#
# Reclassifies the specified raster layer using some polygon layers.  Cells within each 
# polygon layer will be reclassified as the supplied new value classes.  The output
# raster will have the same set of fields and the same symbology.
#
# Spatial Analyst Extension is required for running the script.

import logging, os, re
import arcpy
from arcpy.sa import *
from CscCommon import * 

# Indices of tool parameters, which should be consistent with the exact order
# of the parameters defined for the tool
INPUT_RASTER   = 0
CLASS_NAME_FLD = 1
POLYGON_LAYER  = 2
RECLASS_AS     = 3
RECLASS_LIST   = 4
FILL_NODATA    = 5
OUTPUT_RASTER  = 6

EXCLUDED_FIELDS = ['OBJECTID', 'OID', 'ROWID', 'VALUE', 'COUNT']

name = 'Reclass by Polygons'
logger = logging.getLogger('csc.toolkit.ReclassByPolygons')

def setSymbologyForOutputRaster(inputRaster):
    logger.debug("Saving input raster layer as a layer file...")
    layer = os.path.join(getDocumentDir(), getTempName('INPUT_RASTER') + '.lyr')
    logger.debug("Layer file: %s", layer)
    arcpy.SaveToLayerFile_management(inputRaster, layer)
    
    logger.debug("Configuring symbology for output raster...")
    params = arcpy.GetParameterInfo()
    params[OUTPUT_RASTER].symbology = layer
    
    logger.debug("Symbology for output raster configured.")
    return

def joinFieldsFromInputRaster(inputRaster, outputRaster):
    logger.debug("Building list of fields to join...")
    fields = []
    for field in arcpy.ListFields(inputRaster, None):
        if field.name.upper() not in EXCLUDED_FIELDS:
            fields.append(field.name)
            logger.debug("Field added: %s", field.name)
        else:
            logger.debug("Field excluded: %s", field.name)
    logger.debug("Fields to be joined: %s", str(fields))

    if len(fields) > 0:
        logger.debug("Building attribute table for the output raster...")
        arcpy.BuildRasterAttributeTable_management(outputRaster, "NONE")
        logger.debug("Joining fields...")
        arcpy.JoinField_management(outputRaster, 'Value', inputRaster, 'Value', fields)
        addMessage("Fields joined: %s" % str(fields))
    else:
        addMessage("No fields were joined.")
    
def reclassByPolygons(inputRaster, reclassifications, fillNodata, outputRaster):
    """
    Reclassifies the specified input raster using the supplied polygon layers and their new 
    value classes.  All meaningful fields will be copied over from the input raster layer
    to the output raster data set.  Symbology will be copied over as well.
    
    The reclassification rules should be in the format below.  The name of the new class is
    optional. The whole rule may be enclosed in a pair of single quotes.  Multiple rules
    should be separated by '; ' (a semicolon followed by a space).
        'Layer polygon-layer-name as new-value[: new-class]'
    """
    
    logger.debug("Input raster: %s", inputRaster)
    logger.debug("Reclassifications: %s", reclassifications)
    logger.debug("Fill NODATA: %s", str(fillNodata))
    logger.debug("Output raster: %s", outputRaster)
    
    addMessage("Setting symbology and additional fields for output raster...")
    setSymbologyForOutputRaster(inputRaster)
    
    # Sets the output to the input raster for now
    # For each reclassification, its input raster will be the previous reclassified raster.
    reclassOutput = Raster(inputRaster)
    pattern = re.compile("^\\s*'?Layer (.+) as (\d+)(: .*)?'?\\s*$")
    for reclassification in reclassifications.split(";"):
        logger.debug("Reclassification rule: %s", reclassification)
        
        # Parses the reclassification rule (polygon layer and new value)
        match = pattern.match(reclassification)
        if match is None:
            raise InputError("Unknown reclassification rule: " + reclassification)
        polygonLayer = match.group(1)
        reclassAs = int(match.group(2))
        logger.debug("Polygon layer: %s, reclassify as: %d", polygonLayer, reclassAs)
        
        addMessage("Reclassifying cells within '%s' as %d..." % (polygonLayer, reclassAs))
        
        # Copies the polygon layer to a new feature class
        # If some polygons in this layer are selected, only the selected polygons will be copied.
        logger.debug("Copying (selected) polygons as new feature class...")
        mask = getTempName('POLYGONS')
        arcpy.CopyFeatures_management(polygonLayer, mask)

        # Sets the polygon layer as the mask so that we can reclassify cells within the polygons
        logger.debug("Setting the copied polygon feature class as mask...")
        arcpy.env.mask = mask
        
        # Sets up raster processing environment (snapping raster and raster cell size)
        # This will make sure feature layer (mask) conversion won't cause trouble for the 
        # processing of the input raster.
        arcpy.env.snapRaster = inputRaster
        arcpy.env.cellSize = inputRaster
    
        # Creates a raster indicating whether or not the cells are null
        # This will result in a raster with 0/1 values for all cells within the polygons.
        # We will use this raster instead of the reclass output for the reclassification
        # so that all cells with NODATA values will be reclassified as well.
        logger.debug("Creating a raster indicating NODATA (null) cells (actually for all cells inside the polygons)...")
        insideIsNull = IsNull(inputRaster)
        if insideIsNull is None:
            addMessage("Layer ignored because no cells fall within it.", 1)
            continue
        if DEBUG:
            logger.debug("Saving insideIsNull...")
            insideIsNull.save(getTempName('INSIDE_IS_NULL'))
        
        # Clears the mask
        logger.debug("Clearing the mask...")
        arcpy.env.mask = ''
        
        # Performs the reclassification
        # Since all cells within the polygons (non-null cells) will be assigned to the 
        # same value, we can use a simple math instead of a true reclassification call.
        logger.debug("Reclassifying the cells within polygon layer...")
        if fillNodata:
            logger.debug("NODATA cells inside the polygons will be reclassified as well.")
            reclassPolygons = insideIsNull * 0 + reclassAs
        else:
            logger.debug("NODATA cells will not be reclassified.")
            reclassPolygons = (reclassOutput * 0) + (insideIsNull * 0) + reclassAs
        if DEBUG:
            logger.debug("Saving reclassified cells within polygons...")
            reclassPolygons.save(getTempName('RECLASS'))
        
        addMessage("Cells within '%s' reclassified." % polygonLayer)
        
        addMessage("Merging reclassified cells back into raster...")
        
        logger.debug("Creating a raster indicating inside/outside of the polygons...")
        notReclassified = IsNull(reclassPolygons)
        if DEBUG:
            logger.debug("Saving raster indicating cells not reclassified...")
            notReclassified.save(getTempName('NOT_RECLASS'))
        logger.debug("Combining reclassified cells with previous results...")
        reclassOutput = Con(notReclassified, reclassOutput, reclassPolygons)
        if DEBUG:
            logger.debug("Saving combined reclass result...")
            reclassOutput.save(getTempName('COMBINED'))

    # Save the output
    addMessage("Saving output raster...")
    reclassOutput.save(outputRaster)   

    addMessage("Joining fields from the input raster layer (except %s)..." % str(EXCLUDED_FIELDS))
    joinFieldsFromInputRaster(inputRaster, outputRaster)

    # Done.
    addMessage("%s finished successfully." % name)
    
if __name__ == "__main__":
    logger.debug("%s executed as a script.", name)
    # Runs analysis with all standard options for logging, temporary workspace, error handling, etc
    tryProcessing(reclassByPolygons, \
                  arcpy.GetParameterAsText(INPUT_RASTER), \
                  arcpy.GetParameterAsText(RECLASS_LIST), \
                  ('true' == arcpy.GetParameterAsText(FILL_NODATA).lower()), \
                  arcpy.GetParameterAsText(OUTPUT_RASTER), \
                  name=name, \
                  docdir='ISAT')
        
else:
    logger.debug("%s imported as a lib.", name)
